import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration ;
import java.util.Hashtable ;

/**
 * Defines clusters objects representing the clusters in
 * a populated lattice
 * @author Damien Deville
 *
 */
public class Clusters
{
	private final boolean OCCUP = true ;			// defines an occupied cell
	private final boolean EMPTY = false ;			// defines an empty cell
	private Lattice lattice ;						// defines a lattice
	private Hashtable<Integer,Integer> labels ;		// contains the various labels for the clusters
	private int labelIndex ;						// value of the first label
	private Hashtable<Integer,Integer> clusters ;	// contains the number of cells for each cluster
	
	
	/**
	 * Constructor definition
	 * @param lattice: the lattice we are analysing clusters on
	 */
	public Clusters(Lattice lat)
	{
		lattice = lat ;
		labelIndex = 1 ;								// we set the value of the first label to be 1
		labels = new Hashtable<Integer,Integer>() ;		// we create a new hashtable for the cluster labels
		clusters = new Hashtable<Integer,Integer>() ;	// we create a new hashtable for the cluster numbers
		assignLabels() ;								// we first assign raw cluster labels to the occupied cells
		updateLabels() ;								// we then link the clusters using the root label for each linked cluster
	}
	
	
	/**
	 * Get the cluster number in the lattice
	 * @return the cluster number in the lattice
	 */
	public int getClusterNumber()
	{
		return clusters.size() ;	// the cluster number
	}
	
	
	/**
	 * Get an array populated with the cluster sizes in the lattice
	 * @return an array of cluster sizes
	 */
	public Integer[] getClusterSizes()
	{
		Collection<Integer> c = clusters.values() ;
		Integer[] clusterSizesArray = (Integer[])c.toArray(new Integer[c.size()]) ;
		return clusterSizesArray ;
	}
	
	
	/**
	 * Print the cluster number
	 */
	public void printClusterNumber()
	{
		System.out.println(clusters.size()) ;
	}
	
	
	/**
	 * Print the size of each cluster in the lattice 
	 */
	public void printClusterSizes()
	{
		Enumeration<Integer> k = clusters.keys() ;
		while (k.hasMoreElements())
		{
			int key = (int)k.nextElement() ;
			System.out.println((int)clusters.get(key));
		}
	}
	
	
	/**
	 * Span the lattice once and assign cluster labels to each occupied cell.
	 * However, some cluster can still be defined by more than one label.
	 */
	private void assignLabels()
	{
		for(int i = 0 ; i < lattice.getSize() ; i++)	// we span the lattice
		{
			for(int j = 0 ; j < lattice.getSize() ; j++)
			{
				Cell cell = lattice.getCell(i,j) ;	// we get the current cell in the lattice
				if(cell.getStatus() == OCCUP)		// we check if the cell is occupied
					checkTopAndLeftNeighbors(cell) ;	// we check its top and left neighbors
			}											// to assign the right label to the cell
		}
	}
	
	
	/**
	 * Span the lattice once and update the labels for the clusters that are still
	 * defined by more than one label.
	 * For each label, it checks if it is the root one and if not, it looks for the root one.
	 */
	private void updateLabels()
	{	
		for(int i = 0 ; i < lattice.getSize() ; i++)	// we span the lattice
		{
			for(int j = 0 ; j < lattice.getSize() ; j++)
			{
				Cell cell = lattice.getCell(i,j) ;	// we get the current cell in the lattice
				if(cell.getStatus() == OCCUP)		// we check if the cell is occupied
				{
					int label = findRootLabel(cell) ;	// we find the root label corresponding to the cell
					cell.setClusterLabel(label) ;		// we set this new label to the cell
					if(!clusters.containsKey(label))	// if the clusters hashtable does not already contain this label
						clusters.put(label, 1) ;		// we add it to it
					else
						clusters.put(label, clusters.get(label) + 1) ;	// if it already contains it, we increment its size by 1
				}
			}
		}
	}
	
	
	/**
	 * For each occupied cell, check the cells on top and left of the current cell.
	 * There are 4 cases:
	 * 1: both cells are empty, we then have to create a new cluster label and apply it to the current cell
	 * 2: only one cell is empty, we then have to apply the label of the cluster the non-empty cell belongs to to the current cell
	 * 3: both cells are occupied and have the same cluster label: we then apply this cluster label to the current cell
	 * 4: both cells are occupied but have different cluster labels: we then apply the smallest label to the current cell and define a union between the labels
	 * @param cell: the current cell we are checking
	 */
	private void checkTopAndLeftNeighbors(Cell cell)
	{
		Cell cell_top ;		// cell at the top of the current cell
		Cell cell_left ;	// cell at the left of the current cell
		
		// we make sure that cell_top is actually not outside of the lattice
		if(cell.getI() > 0)		// if it is actually INSIDE the lattice
			cell_top = lattice.getCell(cell.getI()-1, cell.getJ()) ;	// we get the actual cell
		else					// if it is actually OUTSIDE the lattice
			cell_top = new Cell(-1, -1, EMPTY) ;	// we define it as an empty cell with coordinates (-1,-1)
		
		// we make sure that cell_left is actually not outside of the lattice
		if(cell.getJ() > 0)		// if it is actually INSIDE the lattice
			cell_left = lattice.getCell(cell.getI(), cell.getJ()-1) ;	// we get the actual cell
		else					// if it is actually OUTSIDE the lattice
			cell_left = new Cell(-1, -1, EMPTY) ;	// we define is as an empty cell with coordinates (-1,-1)
		
		int topClusterLabel = cell_top.getClusterLabel() ;		// we get the cluster label to which the cell at the top belongs
		int leftClusterLabel = cell_left.getClusterLabel() ;	// we get the cluster label to which the cell at the left belongs
		
		// 1st case:
		// if both cell_top and cell_left are empty, we have to create a new cluster label
		if(cell_top.getStatus() == EMPTY && cell_left.getStatus() == EMPTY)
		{
			cell.setClusterLabel(labelIndex) ;		// we set the cluster label to be labelIndex
			labels.put(labelIndex, labelIndex) ;	// we add this cluster label to the labels hashtable
													// the key and the value is the same since this label is not linked to another one yet
			labelIndex++ ;							// we increment the labelIndex by 1
		}
		
		// 2nd case:
		// if only one of both cell is empty, we have to set the cell cluster label same as the non-empty cell's one
		else if(cell_top.getStatus() == EMPTY)		// if cell_top is empty
			cell.setClusterLabel(find(leftClusterLabel)) ;		// we find the root label and set it as cell_left's label
		else if(cell_left.getStatus() == EMPTY)		// if cell_left is empty
			cell.setClusterLabel(find(topClusterLabel)) ;		// we find the root label and set it as cell_top's label
		
		// 3rd case:
		// if both cell cluster labels are equal, we set the cell cluster label same as this cluster label
		else if(topClusterLabel == leftClusterLabel)
			cell.setClusterLabel(find(topClusterLabel)) ;		// we find the root label and set it as cell_top and cell_left's label
		
		// 4th case:
		// if cell_top and cell_left belong to different clusters, we set the current cell's label as the smallest label
		// we also have to update the labels relation in the labels hashtable
		// in the label hashtable, if one cluster is defined by only one label, the key entry (label) will be the same as its value
		// if one cluster is defined by more than one label, say 2, the key entry will be the biggest label and the value the smallest
		else
		{
			int smallestLabel = (topClusterLabel < leftClusterLabel) ? topClusterLabel : leftClusterLabel ;	// we look for the smallest label between both
			int biggestLabel = (topClusterLabel < leftClusterLabel) ? leftClusterLabel : topClusterLabel ;	// we look for the biggest label between both
			cell.setClusterLabel(smallestLabel) ;		// we set the smallest label for the current cell
			
			if(smallestLabel < labels.get(biggestLabel))	// if the smallest label is actually smaller than the value corresponding to the biggestLabel key
				labels.put(find(labels.get(biggestLabel)), find(smallestLabel)) ;	// we update the key and label to materialize the link
			else											// if the smallest label is actually bigger than the value corresponding to the biggestLabel key
				labels.put(find(smallestLabel), find(labels.get(biggestLabel))) ;	// we update the key and label to materialize the link
		}
	}
	
	
	/**
	 * Find the smallest "good" label given a "bad" label
	 * @param label: the bad label we want to find the good label for
	 * @return the good label
	 */
	private int find(int label)
	{
		int initLabel = label ;		// we store the original bad label 
		while(labels.get(initLabel) != initLabel)	// while the label is not good, we loop through the links
			initLabel = labels.get(initLabel) ;		
		// the value of initLabel is now the "good" label
		
		while(labels.get(label) != label)	// we loop through the labels from the bad one another time 
		{
			int temp = labels.get(label) ;	// we get the current label at each loop step
			labels.put(label, initLabel) ;	// we set the value of each current label to the "good" label
			label = temp ;
		}
		return initLabel ;		// we return the good label	
	}
	
	
	/**
	 * Find the root label for each cluster label from the labels hashtable.
	 * The root label is defined to be the smallest label among all labels constituing a common cluster.
	 * @param cell: the cell we are checking the cluster label for
	 * @return the root cluster label
	 */
	private int findRootLabel(Cell cell)
	{
		int labelNum = cell.getClusterLabel() ;		// we get the cluster label of the cell
		while(labels.get(labelNum) != labelNum)		// we loop through the labels hashtable
		{											// until we reach the root label
			labelNum = labels.get(labelNum) ;
		}
		return labelNum ;							// we return the root label for this cell
	}
	
	
	/**
	 * Check if percolation (a top-bottom spanning cluster) occurs in the lattice
	 * @return true if there is percolation, false if not
	 */
	public boolean checkPercolation()
	{
		// we check if there is the same label on the top and bottom rows of the lattice
		ArrayList<Integer> top = new ArrayList<Integer>() ;		// arraylist that contains all top-row cells
		ArrayList<Integer> bottom = new ArrayList<Integer>() ;	// arraylist that contains all bottom-row cells
		ArrayList<Integer> left = new ArrayList<Integer>() ;	// arraylist that contains all left-column cells
		ArrayList<Integer> right = new ArrayList<Integer>() ;	// arraylist that contains all right-column cells
		for(int i = 0 ; i < lattice.getSize() ; i++)	// we loop through the columns of the lattice
		{
			Cell topCell = lattice.getCell(0,i) ;						// we get each cell from the top row
			Cell bottomCell = lattice.getCell(lattice.getSize()-1,i) ;	// we get each cell from the bottom row
			if(topCell.getStatus() == OCCUP)				// if the top cell is occupied
				top.add(topCell.getClusterLabel()) ;		// we add it to the top arraylist
			if(bottomCell.getStatus() == OCCUP)				// if the bottom cell is occupied
				bottom.add(bottomCell.getClusterLabel()) ;	// we add it to the bottom arraylist
		}
		top.retainAll(bottom) ;		// we remove from the top arraylist all elements that are not in bottom
		if(top.size() > 0)			// we check if there are still elements in the arraylist
			return true ;
		// if there is not a spanning cluster from top to bottom,
		// we check if there is one from left side to right side
		for(int j = 0 ; j < lattice.getSize() ; j++)	// we loop through the rows of the lattice
		{
			Cell leftCell = lattice.getCell(j,0) ;						// we get each cell from the left column
			Cell rightCell = lattice.getCell(j,lattice.getSize()-1) ;	// we get each cell from the right column
			if(leftCell.getStatus() == OCCUP)				// if the left cell is occupied
				left.add(leftCell.getClusterLabel()) ;		// we add it to the left arraylist
			if(rightCell.getStatus() == OCCUP)				// if the right cell is occupied
				right.add(rightCell.getClusterLabel()) ;	// we add it to the right arraylist
		}
		left.retainAll(right) ;		// we remove from the left arraylist all elements that are not in right
		if(left.size() > 0)			// we check if there are still elements in the arraylist
			return true ;
		return false ;		// if there is no spanning cluster, we eventually return false
	}
}