import java.util.Random;

/**
 * Create a ContBouchaud objects representing as the return
 * for one time step computed from the Cont-Boucaud model
 * @author Damien Deville
 *
 */
public class ContBouchaud
{
	private Random random ;
	private double a ;				// the probability that a cluster be active (buys or sells)
	private double lambda ;			// the scaling factor
	private double priceReturn ;	// the price change given by the model
	
	/**
	 * Constructor definition.
	 * Compute the price return with the Cont-Bouchaud model. First populate a lattice, find
	 * the cluster sizes and then compute the price return.
	 * @param latticeSize: the size of the lattice
	 * @param prob: the probability that a site in the lattice be occupied
	 * @param a: the probability that a cluster be active (buys or sells)
	 * @param lambda: the scaling factor in the Cont-Bouchaud model
	 */
	public ContBouchaud(int latticeSize, double prob, double a, double lambda)
	{
		setProb(a) ;						// we set the active cluster probability
		setLambda(lambda) ;					// we set the scaling factor
		
		double sumBuy = 0 ;		// we assume the sum of buying clusters is initially zero
		double sumSell = 0 ;	// we assume the sum of selling clusters is initially zero
		random = new Random() ;	// we define a new random object
		
		Lattice lat = new Lattice(latticeSize, prob) ;	// we create a new lattice given the parameters
		Clusters clusters = new Clusters(lat) ;				// we check the clusters in this lattice
		Integer[] clusterSizes = clusters.getClusterSizes() ;	// we get the size of each cluster
		
		for(int i = 0 ; i < clusterSizes.length ; i++)	// we loop through the clusters
		{
			double rnd = random.nextDouble() ;	// we get a new uniformly generated random number
			if(rnd < a)	// if the random number is less than a (probability a) 
				sumSell += clusterSizes[i] ;	// we assume the cluster is selling
			else if(rnd > 1-a)	// if the random number is greater than 1-a (probability a)
				sumBuy += clusterSizes[i] ;		// we assume the cluster is buying
			// else the cluster sleeps
		}
		priceReturn = (sumBuy-sumSell)/lambda ;	// the price return is then given by the scaled difference
	}
	
	/**
	 * Constructor definition.
	 * Compute the price return using the Cont-Bouchaud model assuming we already have an array
	 * containing the cluster sizes. This constructor is useful when we want to compute various
	 * time steps from the same lattice pattern (for computation efficiency).
	 * @param latticeSize: the size of the lattice
	 * @param prob: the probability that a site in the lattice be occupied
	 * @param a: the probability that a cluster be active (buys or sells)
	 * @param lambda: the scaling factor in the Cont-Bouchaud model
	 */
	public ContBouchaud(double a, double lambda, Integer[] clusterSizes)
	{
		setProb(a) ;						// we set the active cluster probability
		setLambda(lambda) ;					// we set the scaling factor
		
		double sumBuy = 0 ;		// we assume the sum of buying clusters is initially zero
		double sumSell = 0 ;	// we assume the sum of selling clusters is initially zero
		random = new Random() ;	// we define a new random object
		
		for(int i = 0 ; i < clusterSizes.length ; i++)	// we loop through the clusters
		{
			double rnd = random.nextDouble() ;	// we get a new uniformly generated random number
			if(rnd < a)	// if the random number is less than a (probability a) 
				sumSell += clusterSizes[i] ;	// we assume the cluster is selling
			else if(rnd > 1-a)	// if the random number is greater than 1-a (probability a)
				sumBuy += clusterSizes[i] ;		// we assume the cluster is buying
			// else the cluster sleeps
		}
		priceReturn = (sumBuy-sumSell)/lambda ;	// the price return is then given by the scaled difference
	}
	
	
	/**
	 * Get the price return computed with the Cont-Bouchaud model
	 * @return the price return
	 */
	public double getPriceReturn()
	{
		return priceReturn ;
	}
	
	/**
	 * Set the probability that a cluster be active (buys or sells)
	 * @param a: the probability that a site is active (buys or sells)
	 */
	private void setProb(double a)
	{
		this.a = a ;
	}
	
	/**
	 * Set the scaling factor defined in the Cont-Bouchaud model
	 * @param lambda: the scaling factor
	 */
	private void setLambda(double lambda)
	{
		this.lambda = lambda ;
	}
	
	/**
	 * Get the probability that a cluster be active (buys or sells)
	 * @return the probability that a cluster is active (buys or sells)
	 */
	public double getClusterProb()
	{
		return a ;
	}
	
	/**
	 * Get the scaling factor defined in the Cont-Bouchaud model
	 * @return the scaling factor
	 */
	public double getLambda()
	{
		return lambda ;
	}
}
