/**
 * Implement particular models from the Cont-Bouchaud setting
 * @author Damien Deville
 *
 */
public class Model
{
	private int numSteps ;			// the number of time steps
	private double initialPrice ;	// the initial asset price
	
	/**
	 * Constructor definitio 
	 * @param numSteps: the number of time steps in the model
	 * @param initialPrice: the initial asset price
	 */
	public Model(int numSteps, double initialPrice)
	{
		setNumSteps(numSteps) ;			// we set the number of steps
		setInitialPrice(initialPrice) ;	// we set the initial asset price
	}
	
	/**
	 * Set the number of time steps 
	 * @param numSteps: the number of time steps
	 */
	public void setNumSteps(int numSteps)
	{
		this.numSteps = numSteps ;
	}
	
	/**
	 * Set the initial asset price 
	 * @param initialPrice: the initial asset price
	 */
	public void setInitialPrice(double initialPrice)
	{
		this.initialPrice = initialPrice ;
	}
	
	/**
	 * Generate returns from the Cont-Bouchaud model assuming the
	 * lattice does not change at each time step (more efficient)
	 * and that we already have this lattice (inputed in the method) 
	 * @param a: the activity probability
	 * @param lambda: the price scaling factor
	 * @param clusterSizes: array containing cluster sizes
	 * @return array containing returns from the model
	 */
	public double[] classicConstantContBouchaud(double a, double lambda, Integer[] clusterSizes)
	{
		double[] returns = new double[numSteps] ;	// we define an array to contain the returns
		ContBouchaud priceReturn ;	// ContBouchaud object that models one return from the model
		for(int i = 0 ; i < numSteps ; i++)	// we loop as many times as the number of steps
		{
			// get a price return from the Cont-Bouchaud model
			priceReturn = new ContBouchaud(a, lambda, clusterSizes) ;
			returns[i] = priceReturn.getPriceReturn() ;
		}
		return returns ;
	}
	
	/**
	 * Generate returns from the Cont-Bouchaud model assuming the
	 * lattice does not change at each time step (more efficient)
	 * @param latticeSize: the size of the lattice
	 * @param prob: the occupancy probability
	 * @param a: the activity probability
	 * @param lambda: the price scaling factor
	 * @return array containing returns from the model
	 */
	public double[] classicConstantContBouchaud(int latticeSize, double prob, double a, double lambda)
	{
		Lattice lat = new Lattice(latticeSize, prob) ;	// define a new lattice
		Clusters clusters = new Clusters(lat) ;	// we check the clusters in this lattice
		Integer[] clusterSizes = clusters.getClusterSizes() ;	// we get the size of each cluster
		
		double[] returns = new double[numSteps] ;	// we define an array to contain the returns
		ContBouchaud priceReturn ;	// ContBouchaud object that models one return from the model
		for(int i = 0 ; i < numSteps ; i++)	// we loop as many times as the number of steps
		{
			// get a price return from the Cont-Bouchaud model
			priceReturn = new ContBouchaud(a, lambda, clusterSizes) ;
			returns[i] = priceReturn.getPriceReturn() ;
		}
		return returns ;
	}
	
	/**
	 * Generate returns from the Cont-Bouchaud model assuming the
	 * lattice changes at each time step
	 * @param latticeSize
	 * @param prob
	 * @param a
	 * @param lambda
	 * @return
	 */
	public double[] classicContBouchaud(int latticeSize, double prob, double a, double lambda)
	{
		double[] returns = new double[numSteps] ;	// we define an array to contain the returns
		ContBouchaud priceReturn ;	// ContBouchaud object that models one return from the model
		for(int i = 0 ; i < numSteps ; i++)	// we loop as many times as the number of steps
		{
			// get a price return from the Cont-Bouchaud model
			priceReturn = new ContBouchaud(latticeSize, prob, a, lambda) ;
			returns[i] = priceReturn.getPriceReturn() ;
		}
		return returns ;
	}
	
	/**
	 * Generate the price time series from an initial price based on the returns
	 * @param initialPrice: the initial price
	 * @return the price time series
	 */
	public double[] generateTimeSeries(double[] priceReturns)
	{
		double[] timeSeries = new double[numSteps] ;	// we define an array to contain the prices
		timeSeries[0] = initialPrice ;	// the first value of the time-series is the initial price
		for(int i = 0 ; i < numSteps-1 ; i++)	// we loop as many times as the number of steps
			timeSeries[i+1] = timeSeries[i] + priceReturns[i] ;	// generate prices
		return timeSeries ;
	}	
}
